# -*- coding: utf-8 -*-
# @Time    : 04/09/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : evaluation.py
# @Software: PyCharm

class silhouette_coefficient:
    pass

class index_davis_buldin:
    pass

